from collections import namedtuple

IndoorBikeData = namedtuple(
    "IndoorBikeData",
    [
        "instant_speed", # km/h
        "average_speed", # km/h
        "instant_cadence", # rpm
        "average_cadence", # rpm
        "total_distance", # m
        "resistance_level", # unitless
        "instant_power", # W
        "average_power", # W
        "total_energy", # kcal
        "energy_per_hour", # kcal/h
        "energy_per_minute", # kcal/min
        "heart_rate", # bpm
        "metabolic_equivalent", # unitless; metas
        "elapsed_time", # s
        "remaining_time", # s
    ]
)
def parse_indoor_bike_data(message) -> IndoorBikeData:
    flag_more_data = bool(message[0] & 0b00000001)
    flag_average_speed = bool(message[0] & 0b00000010)
    # ANOMALY: In the Bluetooth SIG spec, instantaneous_cadence is reversed (0
    # means present, 1 means not present). The Huawei docs do not mention this
    # reversal. In practice, the Huawei docs seem correct. There is no
    # reversal.
    flag_instantaneous_cadence = bool(message[0] & 0b00000100)
    flag_average_cadence = bool(message[0] & 0b00001000)
    flag_total_distance = bool(message[0] & 0b00010000)
    flag_resistance_level = bool(message[0] & 0b00100000)
    flag_instantaneous_power = bool(message[0] & 0b01000000)
    flag_average_power = bool(message[0] & 0b10000000)
    flag_expended_energy = bool(message[1] & 0b00000001)
    flag_heart_rate = bool(message[1] & 0b00000010)
    flag_metabolic_equivalent = bool(message[1] & 0b00000100)
    flag_elapsed_time = bool(message[1] & 0b00001000)
    flag_remaining_time = bool(message[1] & 0b00010000)

    instant_speed = None
    average_speed = None
    instant_cadence = None
    average_cadence = None
    total_distance = None
    resistance_level = None
    instant_power = None
    average_power = None
    total_energy = None
    energy_per_hour = None
    energy_per_minute = None
    heart_rate = None
    metabolic_equivalent = None
    elapsed_time = None
    remaining_time = None

    # for each True flag, get the value
    # confusingly, the order of the flags is not the same as the order of the
    # values
    i = 2  # start after the flags
    if flag_more_data == 0:
        instant_speed = int.from_bytes(message[i : i + 2], "little", signed=False)/100
        i += 2
    if flag_average_speed:
        average_speed = int.from_bytes(message[i : i + 2], "little", signed=False)/100
        i += 2
    if flag_instantaneous_cadence:
        instant_cadence = int.from_bytes(message[i : i + 2], "little", signed=False) / 2
        i += 2
    if flag_average_cadence:
        average_cadence = int.from_bytes(message[i : i + 2], "little", signed=False) / 2
        i += 2
    if flag_total_distance:
        total_distance = int.from_bytes(message[i : i + 3], "little", signed=False)
        i += 3
    if flag_resistance_level:
        resistance_level = int.from_bytes(message[i : i + 2], "little", signed=True)
        i += 2
    if flag_instantaneous_power:
        instant_power = int.from_bytes(message[i : i + 2], "little", signed=True)
        i += 2
    if flag_average_power:
        average_power = int.from_bytes(message[i : i + 2], "little", signed=True)
        i += 2
    if flag_expended_energy:
        total_energy = int.from_bytes(message[i : i + 2], "little", signed=False)
        energy_per_hour = int.from_bytes(message[i + 2 : i + 4], "little", signed=False)
        energy_per_minute = int.from_bytes(
            message[i + 4 : i + 5], "little", signed=False
        )
        i += 5
    if flag_heart_rate:
        heart_rate = int.from_bytes(message[i : i + 1], "little", signed=False)
        i += 1
    if flag_metabolic_equivalent:
        metabolic_equivalent = int.from_bytes(
            message[i : i + 1], "little", signed=False
        ) / 10
        i += 1
    if flag_elapsed_time:
        elapsed_time = int.from_bytes(message[i : i + 2], "little", signed=False)
        i += 2
    if flag_remaining_time:
        remaining_time = int.from_bytes(message[i : i + 2], "little", signed=False)
        i += 2

    return IndoorBikeData(
        instant_speed,
        average_speed,
        instant_cadence,
        average_cadence,
        total_distance,
        resistance_level,
        instant_power,
        average_power,
        total_energy,
        energy_per_hour,
        energy_per_minute,
        heart_rate,
        metabolic_equivalent,
        elapsed_time,
        remaining_time,
    )



import sys
import asyncio
import struct
import importlib.resources
import pycycling.data

sterzo_measurement_id = '347b0030-7635-408b-8918-8ff3949ce592'
sterzo_control_point_id = '347b0031-7635-408b-8918-8ff3949ce592'
sterzo_challenge_code_id = '347b0032-7635-408b-8918-8ff3949ce592'


class Sterzo:
    def __init__(self, client):
        self._client = client
        self._steering_measurement_callback = None
        self._latest_challenge = None

    async def enable_steering_measurement_notifications(self):
        await self._client.start_notify(sterzo_challenge_code_id, self._challenge_code_indication_handler)
        await self._client.start_notify(sterzo_measurement_id, self._steering_measurement_notification_handler)
        await self._client.write_gatt_char(sterzo_control_point_id, bytearray([0x03, 0x10]))
        while self._latest_challenge is None:
            await asyncio.sleep(2)
        await self._activate_steering_measurements()

    async def _activate_steering_measurements(self):
        # importlib.resources.path is deprecated since 3.11
        if sys.version_info >= (3, 11):
            challenge_file = importlib.resources.files(pycycling.data).joinpath('sterzo-challenge-codes.dat').open('rb')
        else:  # legacy support < 3.9
            challenge_file = importlib.resources.open_binary(pycycling.data, 'sterzo-challenge-codes.dat')

        with challenge_file:
            challenge_file.seek(self._latest_challenge * 2, 1)
            code_1 = int.from_bytes(challenge_file.read(1), 'little')
            code_2 = int.from_bytes(challenge_file.read(1), 'little')

        byte_array = bytearray([0x03, 0x11, code_1, code_2])
        await self._client.write_gatt_char(sterzo_control_point_id, byte_array)
        await asyncio.sleep(1)
        await self._client.write_gatt_char(sterzo_control_point_id, bytearray([0x02, 0x02]))

    async def disable_steering_measurement_notifications(self):
        await self._client.stop_notify(sterzo_measurement_id)

    def set_steering_measurement_callback(self, callback):
        self._steering_measurement_callback = callback

    def _challenge_code_indication_handler(self, sender, data):  # pylint: disable=unused-argument
        self._latest_challenge = int.from_bytes(data[2:4], 'big')

    def _steering_measurement_notification_handler(self, sender, data):  # pylint: disable=unused-argument
        [steering_angle] = struct.unpack('<f', data)
        self._steering_measurement_callback(steering_angle)



# Re-export functions from modules in the ftms_parsers directory
#
from pycycling.ftms_parsers.control_point import *
from pycycling.ftms_parsers.fitness_machine_status import *
from pycycling.ftms_parsers.fitness_machine_feature import *
from pycycling.ftms_parsers.indoor_bike_data import *
from pycycling.ftms_parsers.training_status import *



"""
Interact with a Fitness Machine Service (FTMS) Bluetooth LE device.

Example
=======
This example demonstrates all cycling-related functionalities of a FTMS indoor cycling device.

Please see also information on :ref:`obtaining the Bluetooth address of your device <obtaining_device_address>`.

First, it prints all 'read' characteristics:
    + Supported resistance level range
    + Supported power range
    + Fitness machine features

Then, it starts 'notify' characteristics, which stream data from the device:
    + Indoor bike data (speed, cadence, distance, resistance level, power, time)

Finally, it modifies 'write' characteristics with some time in between:
    + Resistance level
    + Target power (automatically adjusts resistance level based on cadence to maintain same power)

.. literalinclude:: ../examples/fitness_machine_service_example.py

"""
from collections import namedtuple

from pycycling.ftms_parsers import (
    parse_fitness_machine_status,
    parse_indoor_bike_data,
    parse_fitness_machine_feature,
    parse_training_status,
    parse_control_point_response,
    form_ftms_control_command,
    FTMSControlPointOpCode,
    FitnessMachineFeature,
)

# read: Supported Resistance Level Range
ftms_supported_resistance_level_range_characteristic_id = (
    "00002ad6-0000-1000-8000-00805f9b34fb"
)
# read: Supported Power Range
ftms_supported_power_range_characteristic_id = "00002ad8-0000-1000-8000-00805f9b34fb"
# (read): Fitness Machine Feature
ftms_fitness_machine_feature_characteristic_id = "00002acc-0000-1000-8000-00805f9b34fb"
# notify: Indoor Bike Data
ftms_indoor_bike_data_characteristic_id = "00002ad2-0000-1000-8000-00805f9b34fb"
# notify: Fitness Machine Status
ftms_fitness_machine_status_characteristic_id = "00002ad3-0000-1000-8000-00805f9b34fb"
# notify: Training Status
ftms_training_status_characteristic_id = "00002ad3-0000-1000-8000-00805f9b34fb"
# (write, indicate): Fitness Machine Control Point
ftms_fitness_machine_control_point_characteristic_id = (
    "00002ad9-0000-1000-8000-00805f9b34fb"
)

SupportedResistanceLevelRange = namedtuple(
    "SupportedResistanceLevelRange",
    ["minimum_resistance", "maximum_resistance", "minimum_increment"],
)


def _parse_supported_resistance_level_range(message: bytearray) -> SupportedResistanceLevelRange:
    minimum_resistance = int.from_bytes(message[0:2], "little")
    maximum_resistance = int.from_bytes(message[2:4], "little")
    minimum_increment = int.from_bytes(message[4:6], "little")
    return SupportedResistanceLevelRange(
        minimum_resistance, maximum_resistance, minimum_increment
    )


SupportedPowerRange = namedtuple(
    "SupportedPowerRange",
    ["minimum_power", "maximum_power", "minimum_increment"],
)


def _parse_supported_power_range(message: bytearray) -> SupportedPowerRange:
    minimum_power = int.from_bytes(message[0:2], "little")
    maximum_power = int.from_bytes(message[2:4], "little")
    minimum_increment = int.from_bytes(message[4:6], "little")
    return SupportedPowerRange(minimum_power, maximum_power, minimum_increment)


class FitnessMachineService:
    def __init__(self, client):
        self._client = client
        self._control_point_response_callback = None
        self._indoor_bike_data_callback = None
        self._fitness_machine_status_callback = None
        self._training_status_callback = None

    # === READ Characteristics ===
    async def get_supported_resistance_level_range(self) -> SupportedResistanceLevelRange:
        message = await self._client.read_gatt_char(
            ftms_supported_resistance_level_range_characteristic_id
        )
        return _parse_supported_resistance_level_range(message)

    async def get_supported_power_range(self) -> SupportedPowerRange:
        message = await self._client.read_gatt_char(
            ftms_supported_power_range_characteristic_id
        )
        return _parse_supported_power_range(message)

    async def get_fitness_machine_feature(self) -> FitnessMachineFeature:
        message = await self._client.read_gatt_char(
            ftms_fitness_machine_feature_characteristic_id
        )
        return parse_fitness_machine_feature(message)

    # === NOTIFY Characteristics ===
    # ====== Indoor Bike Data ======
    async def enable_indoor_bike_data_notify(self) -> None:
        await self._client.start_notify(
            ftms_indoor_bike_data_characteristic_id,
            self._indoor_bike_data_notification_handler,
        )

    async def disable_indoor_bike_data_notify(self):
        await self._client.stop_notify(ftms_indoor_bike_data_characteristic_id)

    def set_indoor_bike_data_handler(self, callback):
        self._indoor_bike_data_callback = callback

    def _indoor_bike_data_notification_handler(
        self, sender, data
    ):  # pylint: disable=unused-argument
        if self._indoor_bike_data_callback is not None:
            self._indoor_bike_data_callback(parse_indoor_bike_data(data))

    # ====== Fitness Machine Status ======
    async def enable_fitness_machine_status_notify(self) -> None:
        await self._client.start_notify(
            ftms_fitness_machine_status_characteristic_id,
            self._fitness_machine_status_notification_handler,
        )

    async def disable_fitness_machine_status_notify(self):
        await self._client.stop_notify(ftms_fitness_machine_status_characteristic_id)

    def set_fitness_machine_status_handler(self, callback):
        self._fitness_machine_status_callback = callback

    def _fitness_machine_status_notification_handler(
        self, sender, data
    ):  # pylint: disable=unused-argument
        if self._fitness_machine_status_callback is not None:
            self._fitness_machine_status_callback(parse_fitness_machine_status(data))

    # ====== Training Status ======
    async def enable_training_status_notify(self) -> None:
        await self._client.start_notify(
            ftms_training_status_characteristic_id,
            self._training_status_notification_handler,
        )

    async def disable_training_status_notify(self):
        await self._client.stop_notify(ftms_training_status_characteristic_id)

    def set_training_status_handler(self, callback):
        self._training_status_callback = callback

    def _training_status_notification_handler(
        self, sender, data
    ):  # pylint: disable=unused-argument
        if self._training_status_callback is not None:
            self._training_status_callback(parse_training_status(data))

    # === WRITE/INDICATE Characteristics ===
    # ====== Fitness Machine Control Point ======
    async def enable_control_point_indicate(self) -> None:
        await self._client.start_notify(
            ftms_fitness_machine_control_point_characteristic_id,
            self._control_point_response_handler,
        )

    async def disable_control_point_indicate(self):
        await self._client.stop_notify(
            ftms_fitness_machine_control_point_characteristic_id
        )

    def set_control_point_response_handler(self, callback):
        self._control_point_response_callback = callback

    def _control_point_response_handler(
        self, sender, data
    ):  # pylint: disable=unused-argument
        if self._control_point_response_callback is not None:
            self._control_point_response_callback(parse_control_point_response(data))

    async def request_control(self) -> None:
        message = form_ftms_control_command(FTMSControlPointOpCode.REQUEST_CONTROL)
        await self._client.write_gatt_char(
            ftms_fitness_machine_control_point_characteristic_id, message, False
        )

    async def reset(self) -> None:
        message = form_ftms_control_command(FTMSControlPointOpCode.RESET)
        await self._client.write_gatt_char(
            ftms_fitness_machine_control_point_characteristic_id, message, False
        )

    async def set_target_resistance_level(self, level: int) -> None:
        message = form_ftms_control_command(
            FTMSControlPointOpCode.SET_TARGET_RESISTANCE_LEVEL, int(level)
        )
        await self._client.write_gatt_char(
            ftms_fitness_machine_control_point_characteristic_id, message, False
        )

    async def set_target_power(self, power: int) -> None:
        message = form_ftms_control_command(
            FTMSControlPointOpCode.SET_TARGET_POWER, int(power)
        )
        await self._client.write_gatt_char(
            ftms_fitness_machine_control_point_characteristic_id, message, False
        )



import unittest

from pycycling.rear_view_radar import _parse_radar_measurement, RadarMeasurement


class TestRearViewRadar(unittest.TestCase):
    def test__parse_radar_measurement(self):
        self.assertEqual(_parse_radar_measurement(
            bytearray(b'\x12\x83\x03\x1f\x8f\x1c-\x97=G')
            ),
                [
                RadarMeasurement(
                    threat_id = 131, # byte 1+3i
                    distance = 3, # byte 2+3i: in meters
                    speed = 31, # byte 3+3i: in km/h
                    ),
                RadarMeasurement( # different car
                    threat_id = 143, # as identified by different threat_id
                    distance = 28,
                    speed = 45,
                    ),
                RadarMeasurement( # third car
                    threat_id = 151,
                    distance = 61,
                    speed = 71
                    ),
                ]
            )

if __name__ == '__main__':
    unittest.main()



import unittest

from pycycling.battery_service import _parse_battery_level


class TestBatteryService(unittest.TestCase):
    def test__parse_battery_level(self):
        self.assertEqual(_parse_battery_level(bytearray([0])), 0)
        self.assertEqual(_parse_battery_level(bytearray([42])), 42)
        self.assertEqual(_parse_battery_level(bytearray([100])), 100)


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple

heart_rate_measurement_characteristic_id = '00002a37-0000-1000-8000-00805f9b34fb'

HeartRateMeasurement = namedtuple('HeartRateMeasurement', ['sensor_contact', 'bpm', 'rr_interval', 'energy_expended'])


def _parse_hr_measurement(data):
    flags = data[0]

    is_uint16_measurement_mask = 0x01
    is_contact_detected_mask = 0x06
    is_energy_expended_present_mask = 0x08
    is_rr_interval_present_mask = 0x10

    sensor_contact = None
    bpm = None
    rr_interval = []
    energy_expended = None

    measurement_byte_offset = 1
    sensor_contact = bool(flags & is_contact_detected_mask)

    if flags & is_uint16_measurement_mask:
        bpm = int.from_bytes(data[measurement_byte_offset:measurement_byte_offset + 2], 'little')
        measurement_byte_offset += 2
    else:
        bpm = data[measurement_byte_offset]
        measurement_byte_offset += 1

    if flags & is_energy_expended_present_mask:
        energy_expended = int.from_bytes(data[measurement_byte_offset:measurement_byte_offset + 2], 'little')
        measurement_byte_offset += 2

    if flags & is_rr_interval_present_mask:
        while len(data[measurement_byte_offset:]) >= 2:
            rr_interval.append(int.from_bytes(data[measurement_byte_offset:measurement_byte_offset + 2], 'little'))
            measurement_byte_offset += 2

    return HeartRateMeasurement(sensor_contact=sensor_contact,
                                bpm=bpm,
                                rr_interval=rr_interval,
                                energy_expended=energy_expended)


class HeartRateService:
    def __init__(self, client):
        self._client = client
        self._hr_measurement_callback = None

    async def enable_hr_measurement_notifications(self):
        await self._client.start_notify(heart_rate_measurement_characteristic_id,
                                        self._hr_measurement_notification_handler)

    async def disable_hr_measurement_notifications(self):
        await self._client.stop_notify(heart_rate_measurement_characteristic_id)

    def set_hr_measurement_handler(self, callback):
        self._hr_measurement_callback = callback

    def _hr_measurement_notification_handler(self, sender, data):  # pylint: disable=unused-argument
        if self._hr_measurement_callback is not None:
            self._hr_measurement_callback(_parse_hr_measurement(data))



from collections import namedtuple

FitnessMachineFeature = namedtuple(
    "FitnessMachineFeature",
    [
        "avg_speed_supported",
        "cadence_supported",
        "total_distance_supported",
        "inclination_supported",
        "elevation_gain_supported",
        "pace_supported",
        "step_count_supported",
        "resistance_level_supported",
        "stride_count_supported",
        "expended_energy_supported",
        "heart_rate_measurement_supported",
        "metabolic_equivalent_supported",
        "elapsed_time_supported",
        "remaining_time_supported",
        "power_measurement_supported",
        "force_on_belt_and_power_output_supported",
        "user_data_retention_supported",
    ],
)


def parse_fitness_machine_feature(message: bytearray) -> FitnessMachineFeature:
    """Bit flags are set across two message"""
    avg_speed_supported = bool(message[0] & 0b00000001)
    cadence_supported = bool(message[0] & 0b00000010)
    total_distance_supported = bool(message[0] & 0b00000100)
    inclination_supported = bool(message[0] & 0b00001000)
    elevation_gain_supported = bool(message[0] & 0b00010000)
    pace_supported = bool(message[0] & 0b00100000)
    step_count_supported = bool(message[0] & 0b01000000)
    resistance_level_supported = bool(message[0] & 0b10000000)
    stride_count_supported = bool(message[1] & 0b00000001)
    expended_energy_supported = bool(message[1] & 0b00000010)
    heart_rate_measurement_supported = bool(message[1] & 0b00000100)
    metabolic_equivalent_supported = bool(message[1] & 0b00001000)
    elapsed_time_supported = bool(message[1] & 0b00010000)
    remaining_time_supported = bool(message[1] & 0b00100000)
    power_measurement_supported = bool(message[1] & 0b01000000)
    force_on_belt_and_power_output_supported = bool(message[1] & 0b10000000)
    user_data_retention_supported = bool(message[2] & 0b00000001)
    return FitnessMachineFeature(
        avg_speed_supported,
        cadence_supported,
        total_distance_supported,
        inclination_supported,
        elevation_gain_supported,
        pace_supported,
        step_count_supported,
        resistance_level_supported,
        stride_count_supported,
        expended_energy_supported,
        heart_rate_measurement_supported,
        metabolic_equivalent_supported,
        elapsed_time_supported,
        remaining_time_supported,
        power_measurement_supported,
        force_on_belt_and_power_output_supported,
        user_data_retention_supported,
    )



import unittest

from pycycling.heart_rate_service import _parse_hr_measurement, HeartRateMeasurement


class TestHeartRateServiceService(unittest.TestCase):
    def test__parse_csc_measurement(self):
        self.assertEqual(_parse_hr_measurement(
            bytearray([
                0b00000000,  # flags
                0b00101010,  # bpm
            ])),
            HeartRateMeasurement(
                sensor_contact=False,
                bpm=42,
                rr_interval=[],
                energy_expended=None
            )
        )


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple
from enum import Enum


class TrainingStatus(Enum):
    OTHER = 0
    IDLE = 1
    WARMING_UP = 2
    LOW_INTENSITY_INTERVAL = 3
    HIGH_INTENSITY_INTERVAL = 4
    RECOVERY_INTERVAL = 5
    ISOMETRIC = 6
    HEART_RATE_CONTROL = 7
    FITNESS_TEST = 8
    SPEED_OUTSIDE_CONTROL_REGION_LOW = 9
    SPEED_OUTSIDE_CONTROL_REGION_HIGH = 10
    COOL_DOWN = 11
    WATT_CONTROL = 12
    MANUAL_MODE = 13
    PRE_WORKOUT = 14
    POST_WORKOUT = 15
    RESERVED = 16


TrainingStatusMessage = namedtuple(
    "TrainingStatusMessage",
    [
        "param",
        "string",
    ],
)


def parse_training_status(message: bytearray) -> TrainingStatusMessage:
    param = None
    string = None

    param_exists = message[0] & 0b00000001
    string_exists = message[0] & 0b00000010

    if string_exists:
        string = message[2].decode("utf-8")

    if param_exists:
        ts_byte = message[1]
        if ts_byte == 0x00:
            param = TrainingStatus.OTHER
        elif ts_byte == 0x01:
            param = TrainingStatus.IDLE
        elif ts_byte == 0x02:
            param = TrainingStatus.WARMING_UP
        elif ts_byte == 0x03:
            param = TrainingStatus.LOW_INTENSITY_INTERVAL
        elif ts_byte == 0x04:
            param = TrainingStatus.HIGH_INTENSITY_INTERVAL
        elif ts_byte == 0x05:
            param = TrainingStatus.RECOVERY_INTERVAL
        elif ts_byte == 0x06:
            param = TrainingStatus.ISOMETRIC
        elif ts_byte == 0x07:
            param = TrainingStatus.HEART_RATE_CONTROL
        elif ts_byte == 0x08:
            param = TrainingStatus.FITNESS_TEST
        elif ts_byte == 0x09:
            param = TrainingStatus.SPEED_OUTSIDE_CONTROL_REGION_LOW
        elif ts_byte == 0x0A:
            param = TrainingStatus.SPEED_OUTSIDE_CONTROL_REGION_HIGH
        elif ts_byte == 0x0B:
            param = TrainingStatus.COOL_DOWN
        elif ts_byte == 0x0C:
            param = TrainingStatus.WATT_CONTROL
        elif ts_byte == 0x0D:
            param = TrainingStatus.MANUAL_MODE
        elif ts_byte == 0x0E:
            param = TrainingStatus.PRE_WORKOUT
        elif ts_byte == 0x0F:
            param = TrainingStatus.POST_WORKOUT
        elif ts_byte == 0x10:
            param = TrainingStatus.RESERVED
    return TrainingStatusMessage(param, string)



"""
A module for interacting with Bluetooth devices which support the Cycling Power Service.

This service is supported by most power meters and some turbo trainers.

Example
=======
This example prints cycling power measurements broadcast from the Bluetooth device to the console. Please see also
information on :ref:`obtaining the Bluetooth address of your device <obtaining_device_address>`.

.. literalinclude:: ../examples/cycling_power_service_example.py
"""
from collections import namedtuple
from enum import Enum

cycling_power_measurement_tx_id = '00002a63-0000-1000-8000-00805f9b34fb'
cycling_power_vector_tx_id = '00002a64-0000-1000-8000-00805f9b34fb'
cycling_power_feature_tx_id = '00002a65-0000-1000-8000-00805f9b34fb'
sensor_location_tx_id = '00002a5d-0000-1000-8000-00805f9b34fb'

SensorMeasurementContext = Enum('SensorMeasurementContext', 'force_based torque_based')

DistributeSystemSupport = Enum('DistributeSystemSupport',
                               'unspecified no_distributed_system_support distributed_system_support rfu')

SensorLocation = Enum('SensorLocation',
                      'other top_of_shoe in_shoe hip front_wheel left_crank right_crank left_pedal right_pedal '
                      'front_hub rear_dropout chainstay rear_wheel rear_hub chest spider chain_ring')

InstantaneousMeasurementDirection = Enum('InstantaneousMeasurementDirection',
                                         'unknown tangential_component radial_component lateral_component')

CyclingPowerMeasurement = namedtuple('CyclingPowerMeasurement',
                                     ['instantaneous_power', 'accumulated_energy', 'pedal_power_balance',
                                      'accumulated_torque', 'cumulative_wheel_revs', 'last_wheel_event_time',
                                      'cumulative_crank_revs', 'last_crank_event_time', 'maximum_force_magnitude',
                                      'minimum_force_magnitude', 'maximum_torque_magnitude', 'minimum_torque_magnitude',
                                      'top_dead_spot_angle', 'bottom_dead_spot_angle'])

CyclingPowerFeature = namedtuple('CyclingPowerFeature',
                                 ['pedal_power_balance_supported', 'accumulated_torque_supported',
                                  'wheel_rev_supported', 'crank_rev_supported', 'extreme_magnitudes_supported',
                                  'dead_spot_angles_supported', 'accumulated_energy_supported',
                                  'offset_compensation_supported',
                                  'cycling_power_measurement_content_masking_supported', 'multiple_locations_supported',
                                  'crank_length_adjustment_supported', 'chain_length_adjustment_supported',
                                  'chain_weight_adjustment_supported', 'span_length_adjustment_supported',
                                  'sensor_measurement_context', 'instantaneous_measurement_direction_supported',
                                  'factory_calibration_date_supported', 'enhanced_offset_compensation_supported',
                                  'distribute_system_support'])

CyclingPowerVector = namedtuple('CyclingPowerVector',
                                ['instantaneous_measurement_direction', 'cumulative_crank_revs',
                                 'last_crank_event_time', 'first_crank_measurement_angle',
                                 'instantaneous_force_magnitudes', 'instantaneous_torque_magnitudes'])


def _parse_sensor_location(measurement):
    value = int.from_bytes(measurement, 'little')

    if value >= len(SensorLocation):
        return None
    else:
        return SensorLocation(value + 1)


def _parse_cycling_power_feature(measurement):
    value = int.from_bytes(measurement, byteorder='little')

    pedal_power_balance_supported = bool(value & (1))
    accumulated_torque_supported = bool(value & (1 << 1))
    wheel_rev_supported = bool(value & (1 << 2))
    crank_rev_supported = bool(value & (1 << 3))
    extreme_magnitudes_supported = bool(value & (1 << 4))
    dead_spot_angles_supported = bool(value & (1 << 5))
    accumulated_energy_supported = bool(value & (1 << 6))
    offset_compensation_supported = bool(value & (1 << 7))
    cycling_power_measurement_content_masking_supported = bool(value & (1 << 8))
    multiple_locations_supported = bool(value & (1 << 9))
    crank_length_adjustment_supported = bool(value & (1 << 10))
    chain_length_adjustment_supported = bool(value & (1 << 11))
    chain_weight_adjustment_supported = bool(value & (1 << 12))
    span_length_adjustment_supported = bool(value & (1 << 13))
    sensor_measurement_context_value = bool(value & (1 << 14))

    sensor_measurement_context = SensorMeasurementContext.force_based
    if sensor_measurement_context_value:
        sensor_measurement_context = SensorMeasurementContext.torque_based

    instantaneous_measurement_direction_supported = bool(value & (1 << 15))
    factory_calibration_date_supported = bool(value & (1 << 16))
    enhanced_offset_compensation_supported = bool(value & (1 << 17))

    distribute_system_support_value = (value & 0b11000000000000000000) >> 20
    distribute_system_support = DistributeSystemSupport.unspecified
    if distribute_system_support_value == 1:
        distribute_system_support = DistributeSystemSupport.no_distributed_system_support
    elif distribute_system_support_value == 2:
        distribute_system_support = DistributeSystemSupport.distributed_system_support
    elif distribute_system_support_value == 3:
        distribute_system_support = DistributeSystemSupport.rfu

    return CyclingPowerFeature(
        pedal_power_balance_supported=pedal_power_balance_supported,
        accumulated_torque_supported=accumulated_torque_supported,
        wheel_rev_supported=wheel_rev_supported,
        crank_rev_supported=crank_rev_supported,
        extreme_magnitudes_supported=extreme_magnitudes_supported,
        dead_spot_angles_supported=dead_spot_angles_supported,
        accumulated_energy_supported=accumulated_energy_supported,
        offset_compensation_supported=offset_compensation_supported,
        cycling_power_measurement_content_masking_supported=cycling_power_measurement_content_masking_supported,
        multiple_locations_supported=multiple_locations_supported,
        crank_length_adjustment_supported=crank_length_adjustment_supported,
        chain_length_adjustment_supported=chain_length_adjustment_supported,
        chain_weight_adjustment_supported=chain_weight_adjustment_supported,
        span_length_adjustment_supported=span_length_adjustment_supported,
        sensor_measurement_context=sensor_measurement_context,
        instantaneous_measurement_direction_supported=instantaneous_measurement_direction_supported,
        factory_calibration_date_supported=factory_calibration_date_supported,
        enhanced_offset_compensation_supported=enhanced_offset_compensation_supported,
        distribute_system_support=distribute_system_support
    )


def _parse_cycling_power_measurement(data):
    flags = int.from_bytes(data[0:2], 'little')

    pedal_power_balance_included_flag = 1
    pedal_power_balance_reference_flag = 1 << 1  # pylint: disable=unused-variable
    accumulated_torque_present = 1 << 2
    accumulated_torque_source = 1 << 3  # pylint: disable=unused-variable
    wheel_rev_included_flag = 1 << 4
    crank_rev_included_flag = 1 << 5
    extreme_force_included_flag = 1 << 6
    extreme_torque_included_flag = 1 << 7
    extreme_angles_included_flag = 1 << 8
    top_dead_spot_included_flag = 1 << 9
    bottom_dead_spot_included_flag = 1 << 10
    accumulated_energy_included_flag = 1 << 11
    offset_compensation_indicator_flag = 1 << 12  # pylint: disable=unused-variable

    byte_offset = 2

    instantaneous_power = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
    pedal_power_balance = None
    accumulated_torque = None
    cumulative_wheel_revs = None
    last_wheel_event_time = None
    cumulative_crank_revs = None
    last_crank_event_time = None
    maximum_force_magnitude = None
    minimum_force_magnitude = None
    maximum_torque_magnitude = None
    minimum_torque_magnitude = None
    top_dead_spot_angle = None
    bottom_dead_spot_angle = None
    accumulated_energy = None

    byte_offset += 2
    if flags & pedal_power_balance_included_flag:
        pedal_power_balance = data[byte_offset]
        byte_offset += 1

    if flags & accumulated_torque_present:
        accumulated_torque = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & wheel_rev_included_flag:
        cumulative_wheel_revs = int.from_bytes(data[0 + byte_offset:4 + byte_offset], 'little')
        byte_offset += 4
        last_wheel_event_time = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & crank_rev_included_flag:
        cumulative_crank_revs = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2
        last_crank_event_time = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & extreme_force_included_flag:
        maximum_force_magnitude = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2
        minimum_force_magnitude = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & extreme_torque_included_flag:
        maximum_torque_magnitude = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2
        minimum_torque_magnitude = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & extreme_angles_included_flag:
        # TODO: Implement extreme angles
        byte_offset += 3

    if flags & top_dead_spot_included_flag:
        top_dead_spot_angle = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & bottom_dead_spot_included_flag:
        bottom_dead_spot_angle = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if flags & accumulated_energy_included_flag:
        accumulated_energy = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')

    return CyclingPowerMeasurement(instantaneous_power=instantaneous_power,
                                   accumulated_energy=accumulated_energy,
                                   pedal_power_balance=pedal_power_balance,
                                   accumulated_torque=accumulated_torque,
                                   cumulative_wheel_revs=cumulative_wheel_revs,
                                   last_wheel_event_time=last_wheel_event_time,
                                   cumulative_crank_revs=cumulative_crank_revs,
                                   last_crank_event_time=last_crank_event_time,
                                   maximum_force_magnitude=maximum_force_magnitude,
                                   minimum_force_magnitude=minimum_force_magnitude,
                                   maximum_torque_magnitude=maximum_torque_magnitude,
                                   minimum_torque_magnitude=minimum_torque_magnitude,
                                   top_dead_spot_angle=top_dead_spot_angle,
                                   bottom_dead_spot_angle=bottom_dead_spot_angle)


def _parse_cycling_power_vector(data):
    flags = data[0]

    crank_revolutions_present = bool(flags & 0b1)
    first_crank_measurement_angle_present = bool(flags & 0b10)
    instantaneous_force_array_present = bool(flags & 0b100)
    instantaneous_torque_array_present = bool(flags & 0b1000)
    instantaneous_measurement_direction_value = (flags & 0b110000) >> 4

    instantaneous_measurement_direction = InstantaneousMeasurementDirection.unknown
    if instantaneous_measurement_direction_value == 1:
        instantaneous_measurement_direction = InstantaneousMeasurementDirection.tangential_component
    elif instantaneous_measurement_direction_value == 2:
        instantaneous_measurement_direction = InstantaneousMeasurementDirection.radial_component
    elif instantaneous_measurement_direction_value == 3:
        instantaneous_measurement_direction = InstantaneousMeasurementDirection.lateral_component

    byte_offset = 1
    cumulative_crank_revs = None
    last_crank_event_time = None
    first_crank_measurement_angle = None
    instantaneous_force_magnitudes = []
    instantaneous_torque_magnitudes = []

    if crank_revolutions_present:
        cumulative_crank_revs = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2
        last_crank_event_time = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    if first_crank_measurement_angle_present:
        first_crank_measurement_angle = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        byte_offset += 2

    for i in range(byte_offset, len(data), 2):
        element = int.from_bytes(data[i:i + 2], 'little', signed=True)
        if instantaneous_force_array_present:
            instantaneous_force_magnitudes.append(element)
        elif instantaneous_torque_array_present:
            instantaneous_torque_magnitudes.append(element)

    return CyclingPowerVector(instantaneous_measurement_direction=instantaneous_measurement_direction,
                              cumulative_crank_revs=cumulative_crank_revs,
                              last_crank_event_time=last_crank_event_time,
                              first_crank_measurement_angle=first_crank_measurement_angle,
                              instantaneous_force_magnitudes=instantaneous_force_magnitudes,
                              instantaneous_torque_magnitudes=instantaneous_torque_magnitudes)


class CyclingPowerService:
    def __init__(self, client):
        self._client = client
        self._cycling_power_measurement_callback = None
        self._cycling_power_vector_callback = None

    async def enable_cycling_power_measurement_notifications(self):
        await self._client.start_notify(cycling_power_measurement_tx_id,
                                        self._cycling_power_measurement_notification_handler)

    async def disable_cycling_power_measurement_notifications(self):
        await self._client.stop_notify(cycling_power_measurement_tx_id)

    def set_cycling_power_measurement_handler(self, callback):
        self._cycling_power_measurement_callback = callback

    async def enable_cycling_power_vector_notifications(self):
        await self._client.start_notify(cycling_power_vector_tx_id,
                                        self._cycling_power_vector_notification_handler)

    async def disable_cycling_power_vector_notifications(self):
        await self._client.stop_notify(cycling_power_vector_tx_id)

    def set_cycling_power_vector_handler(self, callback):
        self._cycling_power_vector_callback = callback

    async def get_sensor_location(self):
        measurement = await self._client.read_gatt_char(sensor_location_tx_id)
        return _parse_sensor_location(measurement)

    async def get_cycling_power_feature(self):
        measurement = await self._client.read_gatt_char(cycling_power_feature_tx_id)
        return _parse_cycling_power_feature(measurement)

    def _cycling_power_measurement_notification_handler(self, sender, data):  # pylint: disable=unused-argument
        if self._cycling_power_measurement_callback is not None:
            self._cycling_power_measurement_callback(_parse_cycling_power_measurement(data))

    def _cycling_power_vector_notification_handler(self, sender, data):  # pylint: disable=unused-argument
        if self._cycling_power_vector_callback is not None:
            self._cycling_power_vector_callback(_parse_cycling_power_vector(data))



"""
A module for interacting with Tacx Bluetooth smart turbo trainers.

This protocol is a variation of the ANT+ FE-C standard (but using BLE instead of ANT+) - Tacx released many of their
trainers before the now common FTMS protocol was finalised.

The protocol also facilitates the use of the NEO Road Feel feature.

Smart trainer modes of operation
================================
The FE-C standard defines a few modes of operation. To understand the difference between these, a little theory is
required.

Fundamentally, when you cycle two types of forces come into play:

* **Resistive forces**:
    These are typically rolling resistance, wind resistance, and gravitational resistance.
    They are forces that would bring you back to a stop if you stop cycling.
* **Inertial forces**:
    These are the forces that resist changes to your velocity (think Newton's second law). These forces work against you
    when you accelerate making it hard work to get going (and even harder the heavier you are).
    However, when you stop pedalling these forces work in your favour, and help counter the resistive forces which slow
    you down, allowing you to coast along for a while.

Many turbo trainers apply inertial forces through a heavy flywheel, and resistive forces through a brake device. The
Tacx NEO is somewhat unique in that it applies both inertial and resistive forces through electromagnets.

The NEO has 5 different operational modes, which alter the application of these two types of forces. Other trainers
support only the first 3. I'll try and explain in a few words what each does:

* **Basic resistance mode**:
    This is a simple mode, and not useful for many applications. You directly set the resistance force. It applies
    inertial forces assuming a small preset rider mass which means that this mode is not useful for a simulator (because
    it therefore applies the incorrect inertial forces for most riders). In pycycling you activate it by using
    :func:`set_basic_resistance <pycycling.tacx_trainer_control.TacxTrainerControl.set_basic_resistance>`.
* **Simulation mode**:
    Here, the trainer computes the resistive force by internally computing a simple physics equation.
    The equation has parameters which can be configured such as wind speed and track incline. For the full details,
    please refer to the ANT+ FE-C specification. In this mode, the trainer dynamically adjusts the flywheel mass so it
    applies the correct inertial force for the rider. To activate this mode, first use
    :func:`set_user_configuration <pycycling.tacx_trainer_control.TacxTrainerControl.set_user_configuration>` then
    :func:`set_wind_resistance <pycycling.tacx_trainer_control.TacxTrainerControl.set_wind_resistance>` and
    :func:`set_track_resistance <pycycling.tacx_trainer_control.TacxTrainerControl.set_track_resistance>` as required.
    I would strongly recommend using this for any simulator application and this is what is used by Zwift etc.
* **Target power mode (erg mode)**:
    The trainer adjusts the resistance based on the riders exertion in order to maintain a constant power output.
    Use :func:`set_target_power <pycycling.tacx_trainer_control.TacxTrainerControl.set_target_power>` to activate.
* **Isokinetic mode**:
    The trainer adjusts the resistance to maintain a constant cadence.
    See :func:`set_neo_modes <pycycling.tacx_trainer_control.TacxTrainerControl.set_neo_modes>` for info on this.
* **Isotonic mode**:
    This is a special case of basic resistance mode, where the flywheel simulated mass is set to zero rather than a
    small mass, meaning the inertial forces applied by the trainer are zero. To activate, set bike weight, user weight,
    incline, drag resistance and rolling resistance to zero and then use :func:`set_basic_resistance
    <pycycling.tacx_trainer_control.TacxTrainerControl.set_basic_resistance>` as before.

Example
=======
This example demonstrates use of the basic resistance mode. Initially a resistance of 20 newtons is set, and then 20
seconds later this is adjusted to 40 newtons. Data from the trainer is also printed to the console. Please see also
information on :ref:`obtaining the Bluetooth address of your device <obtaining_device_address>`.

.. literalinclude:: ../examples/tacx_trainer_control_example.py
"""
from collections import namedtuple
from enum import Enum

# The GATT Characteristic used for sending FE-C messages to Tacx trainer
tacx_uart_rx_id = '6e40fec3-b5a3-f393-e0a9-e50e24dcca9e'
# The GATT Characteristic used for receiving FE-C messages from Tacx trainer
tacx_uart_tx_id = '6e40fec2-b5a3-f393-e0a9-e50e24dcca9e'

EquipmentType = Enum('EquipmentType', 'treadmill elliptical reserved rower climber nordic_skier trainer')

FEState = Enum('FEState', 'reserved ready in_use finished')

TargetPowerLimit = Enum('TargetPowerLimit',
                        'operating_at_target_or_no_target_set user_speed_too_low user_speed_too_high limit_reached')

CommandStatus = Enum('CommandStatus', 'success fail not_supported rejected uninitialized')


class RoadSurface(Enum):
    """
    Road surfaces supported by the NEO road feel feature
    """
    SIMULATION_OFF = 0
    CONCRETE_PLATES = 1
    CATTLE_GRID = 2
    COBBLESTONES_HARD = 3
    COBBLESTONES_SOFT = 4
    BRICK_ROAD = 5
    OFF_ROAD = 6
    GRAVEL = 7
    ICE = 8
    WOODEN_BOARDS = 9


GeneralFEData = namedtuple('GeneralFEData',
                           ['equipment_type', 'elapsed_time', 'distance_travelled', 'speed', 'heart_rate', 'fe_state',
                            'lap_toggle'])

SpecificTrainerData = namedtuple('SpecificTrainerData',
                                 ['update_event_count', 'instantaneous_cadence', 'accumulated_power',
                                  'instantaneous_power', 'trainer_status', 'target_power_limits', 'fe_state',
                                  'lap_toggle', 'power_calibration_required', 'resistance_calibration_required',
                                  'user_configuration_required'])

CommandStatusData = namedtuple('CommandStatusData', ['last_received_command', 'command_status', 'data'])


class TacxTrainerControl:
    def __init__(self, client):
        self._client = client
        self._general_fe_data_page_callback = None
        self._specific_trainer_data_page_callback = None
        self._command_status_data_page_callback = None

    async def set_basic_resistance(self, resistance):
        """Activate basic resistance mode, with specified resistance

        :param resistance: Resistance to apply to trainer, in newtons
        """
        if resistance < 0 or resistance > 200:
            raise ValueError('resistance must be between 0 and 200')

        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0x30, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF])
        write_value.append(int((resistance / 200) * 200))
        await self._send_fec_cmd(write_value)

    async def set_target_power(self, target_power):
        """Activate target power mode, with specified target power

        :param target_power: Target power, in watts
        """
        if target_power < 0 or target_power > 4000:
            raise ValueError('target_power must be between 0 and 4000')

        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0x31, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF])
        target_power_bytes = int(target_power / 0.25).to_bytes(2, byteorder='little')
        write_value.append(target_power_bytes[0])
        write_value.append(target_power_bytes[1])
        await self._send_fec_cmd(write_value)

    async def set_wind_resistance(self, wind_resistance_coefficient, wind_speed, drafting_factor):
        """Activate simulation mode, specifying wind parameters

        :param wind_resistance_coefficient: Wind resistance coefficient is the product of the frontal surface area,
            drag coefficient and air density of the simulation, in kg/m
        :param wind_speed: Speed of wind acting on cyclist in simulation, in km/h. A positive value represents a head
            wind while a negative value represents a tail wind
        :param drafting_factor: Use parameter to scale wind resistance to simulate drafting behind a virtual opponent
        """
        if wind_resistance_coefficient < 0 or wind_resistance_coefficient > 1.86:
            raise ValueError('wind_resistance_coefficient must be between 0 and 1.86')

        if wind_speed < -127 or wind_speed > 127:
            raise ValueError('wind_speed must be between -127 and 127')

        if drafting_factor < 0 or drafting_factor > 1:
            raise ValueError('drafting_factor must be between 0 and 1')

        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0x32, 0xFF, 0xFF, 0xFF, 0xFF])
        write_value.append(int(wind_resistance_coefficient / 0.01))
        write_value.append(int(wind_speed + 127))
        write_value.append(int(drafting_factor / 0.01))
        await self._send_fec_cmd(write_value)

    async def set_track_resistance(self, grade, coefficient_of_rolling_resistance):
        """Activate simulation mode, specifying track resistance parameters

        :param grade: The grade (slope) of simulated track, in %.
        :param coefficient_of_rolling_resistance: The coefficient of rolling resistance, in dimensionless units
        """
        if grade < -200 or grade > 200:
            raise ValueError('grade must be between -200 and 200')

        if coefficient_of_rolling_resistance < 0 or coefficient_of_rolling_resistance > 0.0127:
            raise ValueError('coefficient_of_rolling_resistance must be between 0 and 0.0127')

        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0x33, 0xFF, 0xFF, 0xFF, 0xFF])
        grade_bytes = int((grade + 200) / 0.01).to_bytes(2, byteorder='little')
        write_value.append(grade_bytes[0])
        write_value.append(grade_bytes[1])
        write_value.append(int(coefficient_of_rolling_resistance / 5e-5))
        await self._send_fec_cmd(write_value)

    async def set_user_configuration(self, user_weight, bicycle_weight,
                                     bicycle_wheel_diameter, gear_ratio):
        """Configure trainer parameters, used when the trainer is in simulation mode

        :param user_weight: Weight of the user in kilograms
        :param bicycle_weight: Weight of bicycle in kilograms
        :param bicycle_wheel_diameter: Diameter of bike wheel, in metres
        :param gear_ratio: The bike gear ratio (front chain ring teeth:rear wheel cog teeth)
        """
        if user_weight < 0 or user_weight > 655.34:
            raise ValueError('user_weight must be between 0 and 655.34')

        if bicycle_weight < 0 or bicycle_weight > 50:
            raise ValueError('bicycle_weight must be between 0 and 50')

        if bicycle_wheel_diameter < 0 or bicycle_wheel_diameter > 2.54:
            raise ValueError('bicycle_wheel_diameter must be between 0 and 2.54')

        if gear_ratio < 0.03 or gear_ratio > 7.65:
            raise ValueError('gear_ratio must be between 0.03 and 7.65')

        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0x37])

        user_weight_bytes = int(user_weight / 0.01).to_bytes(2, byteorder='little')
        write_value.append(user_weight_bytes[0])
        write_value.append(user_weight_bytes[1])
        write_value.append(0xff)
        bicycle_wheel_diameter_offset = int(round((bicycle_wheel_diameter - round(bicycle_wheel_diameter, 2)) / 0.001))
        bicycle_weight_bytes = int(bicycle_weight / 0.05).to_bytes(2, byteorder='little')
        write_value.append(bicycle_wheel_diameter_offset + ((bicycle_weight_bytes[0] << 4) & 0xff))
        write_value.append((bicycle_weight_bytes[0] >> 4) + (bicycle_weight_bytes[1] << 4))
        write_value.append(int(round(bicycle_wheel_diameter, 2) / 0.01))
        write_value.append(int(gear_ratio / 0.03))
        await self._send_fec_cmd(write_value)

    async def set_neo_modes(self, isokinetic_mode=False, isokinetic_speed=4.2,
                            road_surface_pattern=RoadSurface.SIMULATION_OFF,
                            road_surface_pattern_intensity=255):
        """Set NEO specific parameters such as Road Feel mode and Isokinetic training mode

        :param isokinetic_mode: Enable isokinetic mode of the trainer
        :param isokinetic_speed: The target speed used in isokinetic mode
        :param road_surface_pattern: The road surface to be simulated
        :param road_surface_pattern_intensity: The intensity of the feeling of the road surface. Note that even 50%
            feels fairly intense, 100% is untested and may damage the trainer!
        """
        if isokinetic_speed < 4.2 or isokinetic_speed > 8.4:
            raise ValueError('isokinetic_speed must be between 4.2 and 8.4')

        if road_surface_pattern_intensity != 255 and (
                road_surface_pattern_intensity < 0 or road_surface_pattern_intensity > 100):
            raise ValueError('road_surface_pattern_intensity must be between 0 and 100, or set to 255')

        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0xFC, 0x00])
        if isokinetic_mode:
            write_value.append(1)
            write_value.append(int(isokinetic_speed / 0.05))
        else:
            write_value.append(0x00)
            write_value.append(0x00)

        write_value.append(0)
        write_value.append(road_surface_pattern.value)
        write_value.append(road_surface_pattern_intensity)
        write_value.append(0x00)
        await self._send_fec_cmd(write_value)

    async def request_data_page(self, page_number):
        write_value = bytearray([0xA4, 0x09, 0x4F, 0x05, 0x46, 0xFF, 0xFF, 0xFF, 0xFF, 0x80])
        write_value.append(page_number)
        write_value.append(0x01)
        await self._send_fec_cmd(write_value)

    def set_general_fe_data_page_handler(self, callback):
        self._general_fe_data_page_callback = callback

    def set_specific_trainer_data_page_handler(self, callback):
        self._specific_trainer_data_page_callback = callback

    def set_command_status_data_page_handler(self, callback):
        self._command_status_data_page_callback = callback

    async def _send_fec_cmd(self, fec_bytes):
        checksum = sum(fec_bytes[1:]) & 0xFF
        fec_bytes.append(checksum)
        await self._client.write_gatt_char(tacx_uart_rx_id, fec_bytes)

    async def enable_fec_notifications(self):
        await self._client.start_notify(tacx_uart_tx_id, self._fec_notification_handler)

    async def disable_fec_notifications(self):
        await self._client.stop_notify(tacx_uart_tx_id)

    def _fec_notification_handler(self, sender, data):  # pylint: disable=unused-argument
        message_length = data[1]
        message_type = data[2]  # pylint: disable=unused-variable
        message_channel = data[3]  # pylint: disable=unused-variable
        message_data = data[4:4 + message_length - 1]
        data_page_no = message_data[0]

        if data_page_no == 16:
            self._general_fe_data_page_handler(message_data)
        elif data_page_no == 25:
            self._specific_trainer_data_page_handler(message_data)
        elif data_page_no == 71:
            self._command_status_data_page_handler(message_data)

    def _general_fe_data_page_handler(self, message_data):
        equipment_type_code = message_data[1]
        equipment_type = self._equipment_type_from_code(equipment_type_code)

        elapsed_time = message_data[2] * 0.25

        distance_traveled = message_data[3]

        speed_raw = int.from_bytes(message_data[4:6], 'little')
        speed = None
        if speed_raw != 65535:
            speed = speed_raw * 0.001

        heart_rate = message_data[6]
        if heart_rate == 255:
            heart_rate = None

        fe_state, lap_toggle = self._parse_fe_state_nibble((message_data[7] >> 4))

        if self._general_fe_data_page_callback is not None:
            self._general_fe_data_page_callback(GeneralFEData(equipment_type=equipment_type, elapsed_time=elapsed_time,
                                                              distance_travelled=distance_traveled, speed=speed,
                                                              heart_rate=heart_rate, fe_state=fe_state,
                                                              lap_toggle=lap_toggle))

    @staticmethod
    def _equipment_type_from_code(equipment_type_code):
        equipment_type = None
        if equipment_type_code == 19:
            equipment_type = EquipmentType.treadmill
        elif equipment_type_code == 20:
            equipment_type = EquipmentType.elliptical
        elif equipment_type_code == 21:
            equipment_type = EquipmentType.reserved
        elif equipment_type_code == 22:
            equipment_type = EquipmentType.rower
        elif equipment_type_code == 23:
            equipment_type = EquipmentType.climber
        elif equipment_type_code == 24:
            equipment_type = EquipmentType.nordic_skier
        elif equipment_type_code == 25:
            equipment_type = EquipmentType.trainer
        return equipment_type

    @staticmethod
    def _parse_fe_state_nibble(fe_state_nibble):
        lap_toggle = bool(fe_state_nibble & 0x8)
        code = fe_state_nibble & 0x7
        fe_state = None
        if code == 0:
            fe_state = FEState.reserved
        elif code == 1:
            fe_state = FEState.asleep
        elif code == 2:
            fe_state = FEState.ready
        elif code == 3:
            fe_state = FEState.in_use
        elif code == 4:
            fe_state = FEState.finished
        return fe_state, lap_toggle

    def _specific_trainer_data_page_handler(self, message_data):
        update_event_count = message_data[1]

        instantaneous_cadence = message_data[2]
        if instantaneous_cadence == 255:
            instantaneous_cadence = None

        accumulated_power = int.from_bytes(message_data[3:5], 'little')

        power_lsb = message_data[5]
        power_msb = message_data[6]
        instantaneous_power = power_lsb + ((power_msb & 0xf) << 8)

        if instantaneous_power == 4095:
            instantaneous_power = None

        trainer_status_flags = (power_msb >> 4) & 0xf

        power_calibration_required = bool(trainer_status_flags & 0x1)
        resistance_calibration_required = bool(trainer_status_flags & 0x2)
        user_configuration_required = bool(trainer_status_flags & 0x4)

        fe_state, lap_toggle = self._parse_fe_state_nibble((message_data[7] >> 4))
        target_power_limits = None

        flags = message_data[7] & 0x7

        if flags == 0:
            target_power_limits = TargetPowerLimit.operating_at_target_or_no_target_set
        elif flags == 1:
            target_power_limits = TargetPowerLimit.user_speed_too_low
        elif flags == 2:
            target_power_limits = TargetPowerLimit.user_speed_too_high
        elif flags == 3:
            target_power_limits = TargetPowerLimit.limit_reached

        if self._specific_trainer_data_page_callback is not None:
            self._specific_trainer_data_page_callback(
                SpecificTrainerData(update_event_count=update_event_count,
                                    instantaneous_cadence=instantaneous_cadence,
                                    accumulated_power=accumulated_power,
                                    instantaneous_power=instantaneous_power,
                                    trainer_status=None,
                                    target_power_limits=target_power_limits,
                                    fe_state=fe_state, lap_toggle=lap_toggle,
                                    power_calibration_required=power_calibration_required,
                                    resistance_calibration_required=resistance_calibration_required,
                                    user_configuration_required=user_configuration_required))

    def _command_status_data_page_handler(self, message_data):
        last_received_command = message_data[1]
        command_status = None
        command_status_byte = message_data[3]

        if command_status_byte == 0:
            command_status = CommandStatus.success
        elif command_status_byte == 1:
            command_status = CommandStatus.fail
        elif command_status_byte == 2:
            command_status = CommandStatus.not_supported
        elif command_status_byte == 3:
            command_status = CommandStatus.rejected
        elif command_status_byte == 255:
            command_status = CommandStatus.uninitialized

        if self._command_status_data_page_callback is not None:
            self._command_status_data_page_callback(
                CommandStatusData(last_received_command=last_received_command, command_status=command_status,
                                  data=message_data[4:8]))



"""
A module for interacting with Bluetooth devices which support the Battery Service.

Example
=======
This example prints the current battery level to the console. Please see also information on
:ref:`obtaining the Bluetooth address of your device <obtaining_device_address>`.

.. literalinclude:: ../examples/battery_example.py
"""

battery_level_characteristic_id = '00002a19-0000-1000-8000-00805f9b34fb'


def _parse_battery_level(measurement):
    return int.from_bytes(measurement, byteorder='little')


class BatteryService:
    """
    A wrapper around a :obj:`bleak.backends.client.BaseBleakClient` object adding Battery Service specific utility
    methods.

    :param client: A valid :obj:`bleak.backends.client.BaseBleakClient` object
    """

    def __init__(self, client):
        self._client = client

    async def get_battery_level(self):
        """
        Returns current battery level of the Bluetooth device.

        :return: An :obj:`int` representing the current battery level percentage of the Bluetooth device. A value of
            `100` indicates a fully charged battery while `0` a fully discharged battery.
        """
        measurement = await self._client.read_gatt_char(battery_level_characteristic_id)
        return _parse_battery_level(measurement)



import unittest

from pycycling.cycling_speed_cadence_service import _parse_csc_feature, _parse_csc_measurement, CSCFeature, \
    CSCMeasurement


class TestCyclingSpeedCadenceService(unittest.TestCase):
    def test__parse_csc_feature(self):
        self.assertEqual(_parse_csc_feature(
            bytearray([
                0b00000111, 0b00000000
            ])),
            CSCFeature(
                wheel_rev_supported=True,
                crank_rev_supported=True,
                multiple_locations_supported=True
            )
        )

        self.assertEqual(_parse_csc_feature(
            bytearray([
                0b00000101, 0b00000000
            ])),
            CSCFeature(
                wheel_rev_supported=True,
                crank_rev_supported=False,
                multiple_locations_supported=True
            )
        )

        self.assertEqual(_parse_csc_feature(
            bytearray([
                0b00000000, 0b00000000
            ])),
            CSCFeature(
                wheel_rev_supported=False,
                crank_rev_supported=False,
                multiple_locations_supported=False
            )
        )

    def test__parse_csc_measurement(self):
        self.assertEqual(_parse_csc_measurement(
            bytearray([
                0b00000011,  # flags
                0b11111111, 0b11111111, 0b11111111, 0b11111111,  # wheel revs
                0b00000011, 0b00111111,  # last wheel event time
                0b11111111, 0b00000111,  # crank revs
                0b11100111, 0b00001111,  # last crank event time
            ])),
            CSCMeasurement(
                cumulative_wheel_revs=4294967295,
                last_wheel_event_time=16131,
                cumulative_crank_revs=2047,
                last_crank_event_time=4071
            )
        )


if __name__ == '__main__':
    unittest.main()



""" A module for interacting with Bluetooth LE devices which support the Radar (RDR) service

Jason Sohn 2022

This service is tested on Garmin Varia RVR315.
Other models which are expected to support RDR service are:

* Garmin RTR515, RTR516 (German market version), and RCT715
* Bryton Gardia R300
* Magene L508

Example
=======
This example prints radar information broadcast from the Bluetooth device to the console. Please see also
information on :ref:`obtaining the Bluetooth address of your device <obtaining_device_address>`.

.. literalinclude:: ../examples/rear_view_radar_example.py
"""

from collections import namedtuple

radar_characteristic_id = '6a4e3203-667b-11e3-949a-0800200c9a66'

RadarMeasurement = namedtuple('RadarMeasurement', [
    'threat_id',
    'speed',
    'distance',
])

def _parse_radar_measurement(data: bytearray) -> RadarMeasurement:
    """
    Characteristic payload in bytes: 1+3i where i is number of threats (cars)

    byte 0: probably some kind of packet identifier (can be used in case of multiple packets)
    byte 1: threat identifier
    byte 2 (5, 8, ...): distance to threat in meters
    byte 3 (6, 9, ...): speed of threat in km/h

    See source for this reverse-engineering in repo README
    """
    radar_measurements = []
    try:
        for i in range(1, len(data), 3):
            threat_id = int(data[i])
            distance = int(data[i+1])
            speed = int(data[i+2])
            radar_measurements.append(RadarMeasurement(threat_id, speed, distance))
    except IndexError:
        print('pycycling:rear_view_radar.py IndexError: probably starting up and not all data is available yet')
        return None
    return radar_measurements

class RearViewRadarService:
    def __init__(self, client):
        self._client = client
        self._radar_measurement_callback = None

    async def enable_radar_measurement_notifications(self):
        await self._client.start_notify(radar_characteristic_id, self._radar_measurement_notification_handler)

    async def disable_radar_measurement_notifications(self):
        await self._client.stop_notify(radar_characteristic_id)

    def set_radar_measurement_handler(self, callback):
        self._radar_measurement_callback = callback

    def _radar_measurement_notification_handler(self, sender, data): # pylint: disable=unused-argument
        if self._radar_measurement_callback is not None:
            self._radar_measurement_callback(_parse_radar_measurement(data))



"""
pycycling is a cross-platform Python package for interacting with various Bluetooth cycling peripherals.

The package provides a number of classes which wrap around `Bleak <https://github.com/hbldh/bleak>`_ client objects.

.. _obtaining_device_address:

Obtaining the address of your device
====================================

This documentation includes a number of code snippets demonstrating package usage. Each of these has a
hardcoded device address ID which you will need to replace with the address of your device. The following script can
be used to obtain this address:

.. literalinclude:: ../examples/discover_devices_example.py
"""



import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pycycling",
    version="0.3.2",
    author="Zachary Bull",
    author_email="zacharyedwardbull@gmail.com",
    description=
    "A Python package for interacting with Bluetooth Low Energy (BLE) compatible bike trainers and power meters",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zacharyedwardbull/pycycling",
    packages=setuptools.find_packages(),
    install_requires=["bleak"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    include_package_data=True,
    package_data={"": ["data/*.dat"]},
)



import unittest

from pycycling.cycling_power_service import _parse_sensor_location, _parse_cycling_power_feature, \
    _parse_cycling_power_measurement, SensorLocation, CyclingPowerFeature, SensorMeasurementContext, \
    DistributeSystemSupport, CyclingPowerMeasurement


class TestCyclingPowerService(unittest.TestCase):
    def test__parse_sensor_location(self):
        self.assertEqual(_parse_sensor_location(bytearray([0])), SensorLocation.other)
        self.assertEqual(_parse_sensor_location(bytearray([3])), SensorLocation.hip)
        self.assertEqual(_parse_sensor_location(bytearray([5])), SensorLocation.left_crank)
        self.assertEqual(_parse_sensor_location(bytearray([17])), None)

    def test__parse_cycling_power_feature(self):
        # device which supports everything
        self.assertEqual(_parse_cycling_power_feature(
            bytearray([
                0b11111111, 0b11111111, 0b11111111, 0b00000000
            ])),
            CyclingPowerFeature(
                pedal_power_balance_supported=True,
                accumulated_torque_supported=True,
                wheel_rev_supported=True,
                crank_rev_supported=True,
                extreme_magnitudes_supported=True,
                dead_spot_angles_supported=True,
                accumulated_energy_supported=True,
                offset_compensation_supported=True,
                cycling_power_measurement_content_masking_supported=True,
                multiple_locations_supported=True,
                crank_length_adjustment_supported=True,
                chain_length_adjustment_supported=True,
                chain_weight_adjustment_supported=True,
                span_length_adjustment_supported=True,
                sensor_measurement_context=SensorMeasurementContext.torque_based,
                instantaneous_measurement_direction_supported=True,
                factory_calibration_date_supported=True,
                enhanced_offset_compensation_supported=True,
                distribute_system_support=DistributeSystemSupport.unspecified
            )
        )
        # device which supports nothing
        self.assertEqual(_parse_cycling_power_feature(
            bytearray([
                0b00000000, 0b00000000, 0b00000000, 0b00000000
            ])),
            CyclingPowerFeature(
                pedal_power_balance_supported=False,
                accumulated_torque_supported=False,
                wheel_rev_supported=False,
                crank_rev_supported=False,
                extreme_magnitudes_supported=False,
                dead_spot_angles_supported=False,
                accumulated_energy_supported=False,
                offset_compensation_supported=False,
                cycling_power_measurement_content_masking_supported=False,
                multiple_locations_supported=False,
                crank_length_adjustment_supported=False,
                chain_length_adjustment_supported=False,
                chain_weight_adjustment_supported=False,
                span_length_adjustment_supported=False,
                sensor_measurement_context=SensorMeasurementContext.force_based,
                instantaneous_measurement_direction_supported=False,
                factory_calibration_date_supported=False,
                enhanced_offset_compensation_supported=False,
                distribute_system_support=DistributeSystemSupport.unspecified
            )
        )

    def test__parse_cycling_power_measurement(self):
        # Simple case, just power and no bells and whistles
        self.assertEqual(_parse_cycling_power_measurement(
            bytearray([
                0b00000000, 0b00000000,  # flags
                0b11010010, 0b00000100,  # power data
            ])),
            CyclingPowerMeasurement(
                instantaneous_power=1234,
                accumulated_energy=None,
                pedal_power_balance=None,
                accumulated_torque=None,
                cumulative_wheel_revs=None,
                last_wheel_event_time=None,
                cumulative_crank_revs=None,
                last_crank_event_time=None,
                maximum_force_magnitude=None,
                minimum_force_magnitude=None,
                maximum_torque_magnitude=None,
                minimum_torque_magnitude=None,
                top_dead_spot_angle=None,
                bottom_dead_spot_angle=None
            )
        )

        # Include crank data
        self.assertEqual(_parse_cycling_power_measurement(
            bytearray([
                0b00110000, 0b00000000,  # flags
                0b11010010, 0b00000100,  # power data
                0b11111111, 0b11111111, 0b11111111, 0b11111111,  # wheel revs
                0b00000011, 0b00111111,  # last wheel event time
                0b11111111, 0b00000111,  # crank revs
                0b11100111, 0b00001111,  # last crank event time
            ])),
            CyclingPowerMeasurement(
                instantaneous_power=1234,
                accumulated_energy=None,
                pedal_power_balance=None,
                accumulated_torque=None,
                cumulative_wheel_revs=4294967295,
                last_wheel_event_time=16131,
                cumulative_crank_revs=2047,
                last_crank_event_time=4071,
                maximum_force_magnitude=None,
                minimum_force_magnitude=None,
                maximum_torque_magnitude=None,
                minimum_torque_magnitude=None,
                top_dead_spot_angle=None,
                bottom_dead_spot_angle=None
            )
        )


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from collections import namedtuple

FitnessMachineStatusMessage = namedtuple("FitnessMachineStatusMessage", [
    "status",
    "value",
    "unit",
])


class FitnessMachineStatus(Enum):
    RESERVED_FOR_FUTURE_USE = 0
    RESET = 1
    STOPPED_BY_USER = 2
    PAUSED_BY_USER = 3
    STOPPED_BY_SAFETY_KEY = 4
    STARTED_BY_USER = 5
    NEW_SPEED = 6
    NEW_INCLINATION = 7
    NEW_RESISTANCE = 8
    NEW_POWER = 9
    NEW_HEART_RATE = 10
    NEW_EXPENDED_ENERGY = 11
    NEW_NUMBER_OF_STEPS = 12
    NEW_NUMBER_OF_STRIDES = 13
    NEW_DISTANCE = 14
    NEW_TRAINING_TIME = 15
    NEW_TWO_HEART_RATE_ZONE_TARGET_TIME = 16
    NEW_THREE_HEART_RATE_ZONE_TARGET_TIME = 17
    NEW_FIVE_HEART_RATE_ZONE_TARGET_TIME = 18
    NEW_INDOOR_BIKE_SIMULATION_PARAMETERS = 19
    NEW_WHEEL_CIRCUMFERENCE = 20
    NEW_SPIN_DOWN_STATUS = 21
    NEW_TARGET_CADENCE = 22
    CONTROL_PERMISSION_LOST = 23


TwoZoneHR = namedtuple(
    "TwoZoneHR",
    [
        "fat_burn",
        "fitness",
    ],
)

ThreeZoneHR = namedtuple(
    "ThreeZoneHR",
    [
        "very_light",
        "light",
        "moderate",
    ],
)

FiveZoneHR = namedtuple(
    "FiveZoneHR",
    [
        "very_light",
        "light",
        "moderate",
        "hard",
        "maximum",
    ],
)

IndoorBikeSimulationParameters = namedtuple(
    "IndoorBikeSimulationParameters",
    [
        "wind_speed",
        "grade",
        "coefficient_of_rolling_resistance",
        "wind_resistance_coefficient",
    ],
)


class SpinDownStatusValue(Enum):
    RESERVED_FOR_FUTURE_USE = 0x00
    SPIN_DOWN_REQUESTED = 0x01
    SUCCESS = 0x02
    ERROR = 0x03
    STOP_PEDALING = 0x04


def parse_fitness_machine_status(message: bytearray) -> FitnessMachineStatusMessage:
    """
    A tuple with three items:
    1. A FitnessMachineStatus enum
    2. Associated data (dictionary or namedtuple())
    3. Units
    """
    parsed_status = (None, None, None)
    if message[0] == 0x00:
        parsed_status = (FitnessMachineStatus.RESERVED_FOR_FUTURE_USE, None, None)
    elif message[0] == 0x01:
        parsed_status = (FitnessMachineStatus.RESET, None, None)
    elif message[0] == 0x02:
        if message[1] == 0x01:
            parsed_status = (FitnessMachineStatus.STOPPED_BY_USER, None, None)
        elif message[1] == 0x02:
            parsed_status = (FitnessMachineStatus.PAUSED_BY_USER, None, None)
        else:
            parsed_status = (FitnessMachineStatus.RESERVED_FOR_FUTURE_USE, None, None)
    elif message[0] == 0x03:
        parsed_status = (FitnessMachineStatus.STOPPED_BY_SAFETY_KEY, None, None)
    elif message[0] == 0x04:
        parsed_status = (FitnessMachineStatus.STARTED_BY_USER, None, None)
    elif message[0] == 0x05:
        speed = int.from_bytes(message[1:3], byteorder="little", signed=False) / 100
        parsed_status = (FitnessMachineStatus.NEW_SPEED, speed, "km/h")
    elif message[0] == 0x06:
        inclination = int.from_bytes(message[1:3], byteorder="little", signed=True) / 10
        parsed_status = (FitnessMachineStatus.NEW_INCLINATION, inclination, "%")
    elif message[0] == 0x07:
        resistance_level = int.from_bytes(message[1:3], byteorder="little", signed=True)
        parsed_status = (FitnessMachineStatus.NEW_RESISTANCE, resistance_level, "%")
    elif message[0] == 0x08:
        power = int.from_bytes(message[1:3], byteorder="little", signed=True)
        parsed_status = (FitnessMachineStatus.NEW_POWER, power, "W")
    elif message[0] == 0x09:
        heart_rate = int.from_bytes(message[1], byteorder="little", signed=False)
        parsed_status = (FitnessMachineStatus.NEW_HEART_RATE, heart_rate, "bpm")
    elif message[0] == 0x0A:
        expended_energy = int.from_bytes(message[1:3], byteorder="little", signed=False)
        parsed_status = (
            FitnessMachineStatus.NEW_EXPENDED_ENERGY,
            expended_energy,
            "kcal",
        )
    elif message[0] == 0x0B:
        number_of_steps = int.from_bytes(message[1:3], byteorder="little", signed=False)
        parsed_status = (
            FitnessMachineStatus.NEW_NUMBER_OF_STEPS,
            number_of_steps,
            "steps",
        )
    elif message[0] == 0x0C:
        number_of_strides = int.from_bytes(
            message[1:3], byteorder="little", signed=False
        )
        parsed_status = (
            FitnessMachineStatus.NEW_NUMBER_OF_STRIDES,
            number_of_strides,
            "strides",
        )
    elif message[0] == 0x0D:
        distance = int.from_bytes(message[1:4], byteorder="little", signed=False)
        parsed_status = (FitnessMachineStatus.NEW_DISTANCE, distance, "m")
    elif message[0] == 0x0E:
        training_time = int.from_bytes(message[1:3], byteorder="little", signed=False)
        parsed_status = (FitnessMachineStatus.NEW_TRAINING_TIME, training_time, "s")
    elif message[0] == 0x0F:
        parameter = TwoZoneHR(
            fat_burn=int.from_bytes(message[1:3], byteorder="little", signed=False),
            fitness=int.from_bytes(message[3:5], byteorder="little", signed=False),
        )
        parsed_status = (
            FitnessMachineStatus.NEW_TWO_HEART_RATE_ZONE_TARGET_TIME,
            parameter,
            "bpm",
        )
    elif message[0] == 0x10:
        parameter = ThreeZoneHR(
            very_light=int.from_bytes(message[1:3], byteorder="little", signed=False),
            light=int.from_bytes(message[3:5], byteorder="little", signed=False),
            moderate=int.from_bytes(message[5:7], byteorder="little", signed=False),
        )
        parsed_status = (
            FitnessMachineStatus.NEW_THREE_HEART_RATE_ZONE_TARGET_TIME,
            parameter,
            "bpm",
        )
    elif message[0] == 0x11:
        parameter = FiveZoneHR(
            very_light=int.from_bytes(message[1:3], byteorder="little", signed=False),
            light=int.from_bytes(message[3:5], byteorder="little", signed=False),
            moderate=int.from_bytes(message[5:7], byteorder="little", signed=False),
            hard=int.from_bytes(message[7:9], byteorder="little", signed=False),
            maximum=int.from_bytes(message[9:11], byteorder="little", signed=False),
        )
        parsed_status = (
            FitnessMachineStatus.NEW_FIVE_HEART_RATE_ZONE_TARGET_TIME,
            parameter,
            "bpm",
        )
    elif message[0] == 0x12:
        parameter = IndoorBikeSimulationParameters(
            wind_speed=int.from_bytes(message[1:3], byteorder="little", signed=False) / 1000,
            grade=int.from_bytes(message[3:5], byteorder="little", signed=False) / 100,
            coefficient_of_rolling_resistance=int.from_bytes(
                message[5:6], byteorder="little", signed=False
            ) / 1000,
            wind_resistance_coefficient=int.from_bytes(
                message[6:7], byteorder="little", signed=False
            ) / 100,
        )
        parsed_status = (
            FitnessMachineStatus.NEW_INDOOR_BIKE_SIMULATION_PARAMETERS,
            parameter,
            "m/s, %, unitless, kg/m",
        )
    elif message[0] == 0x13:
        wheel_circumference = int.from_bytes(
            message[1:3], byteorder="little", signed=False
        )
        parsed_status = (
            FitnessMachineStatus.NEW_WHEEL_CIRCUMFERENCE,
            wheel_circumference,
            "m",
        )
    elif message[0] == 0x14:
        spin_down_status = int.from_bytes(
            message[1:2], byteorder="little", signed=False
        )
        parsed_status = (
            FitnessMachineStatus.NEW_SPIN_DOWN_STATUS,
            SpinDownStatusValue(spin_down_status),
            "unitless",
        )
    elif message[0] == 0x15:
        target_cadence = int.from_bytes(message[1:3], byteorder="little", signed=False)
        parsed_status = (FitnessMachineStatus.NEW_TARGET_CADENCE, target_cadence, "rpm")
    elif message[0] == 0xFF:
        parsed_status = (FitnessMachineStatus.CONTROL_PERMISSION_LOST, None, None)
    return FitnessMachineStatusMessage(
        status=parsed_status[0],
        value=parsed_status[1],
        unit=parsed_status[2],
    )



from collections import namedtuple

csc_measurement_tx_id = '00002a5b-0000-1000-8000-00805f9b34fb'
csc_feature_tx_id = '00002a5c-0000-1000-8000-00805f9b34fb'

CSCMeasurement = namedtuple('CSCMeasurement',
                            ['cumulative_wheel_revs', 'last_wheel_event_time', 'cumulative_crank_revs',
                             'last_crank_event_time'])

CSCFeature = namedtuple('CSCFeature', ['wheel_rev_supported', 'crank_rev_supported', 'multiple_locations_supported'])


def _parse_csc_feature(measurement):
    value = int.from_bytes(measurement, byteorder='little')
    wheel_rev_supported = bool(value & 0b1)
    crank_rev_supported = bool(value & 0b10)
    multiple_locations_supported = bool(value & 0b100)
    return CSCFeature(wheel_rev_supported=wheel_rev_supported,
                      crank_rev_supported=crank_rev_supported,
                      multiple_locations_supported=multiple_locations_supported)


def _parse_csc_measurement(data):
    flags = data[0]

    wheel_rev_included_flag = 1
    crank_rev_included_flag = 2

    cumulative_wheel_revs = None
    last_wheel_event_time = None
    cumulative_crank_revs = None
    last_crank_event_time = None

    byte_offset = 1
    if flags & wheel_rev_included_flag:
        cumulative_wheel_revs = int.from_bytes(data[0 + byte_offset:4 + byte_offset], 'little')
        last_wheel_event_time = int.from_bytes(data[4 + byte_offset:6 + byte_offset], 'little')
        byte_offset += 6

    if flags & crank_rev_included_flag:
        cumulative_crank_revs = int.from_bytes(data[0 + byte_offset:2 + byte_offset], 'little')
        last_crank_event_time = int.from_bytes(data[2 + byte_offset:4 + byte_offset], 'little')

    return CSCMeasurement(cumulative_wheel_revs=cumulative_wheel_revs,
                          last_wheel_event_time=last_wheel_event_time,
                          cumulative_crank_revs=cumulative_crank_revs,
                          last_crank_event_time=last_crank_event_time)


class CyclingSpeedCadenceService:
    def __init__(self, client):
        self._client = client
        self._csc_measurement_callback = None

    async def enable_csc_measurement_notifications(self):
        await self._client.start_notify(csc_measurement_tx_id, self._csc_measurement_notification_handler)

    async def disable_csc_measurement_notifications(self):
        await self._client.stop_notify(csc_measurement_tx_id)

    def set_csc_measurement_handler(self, callback):
        self._csc_measurement_callback = callback

    async def get_csc_feature(self):
        measurement = await self._client.read_gatt_char(csc_feature_tx_id)
        return _parse_csc_feature(measurement)

    def _csc_measurement_notification_handler(self, sender, data):  # pylint: disable=unused-argument
        if self._csc_measurement_callback is not None:
            self._csc_measurement_callback(_parse_csc_measurement(data))



from enum import Enum
from collections import namedtuple


class FTMSControlPointResponseResultCode(Enum):
    SUCCESS = 0x01
    NOT_SUPPORTED = 0x02
    INCORRECT_PARAMETER = 0x03
    OPERATION_FAILED = 0x04
    CONTROL_NOT_PERMITTED = 0x05


class FTMSControlPointOpCode(Enum):
    REQUEST_CONTROL = 0x00
    RESET = 0x01
    SET_TARGET_SPEED = 0x02
    SET_TARGET_INCLINE = 0x03
    SET_TARGET_RESISTANCE_LEVEL = 0x04
    SET_TARGET_POWER = 0x05
    START_OR_RESUME = 0x06
    STOP_OR_PAUSE = 0x07
    RESPONSE_CODE = 0x80


def form_ftms_control_command(opcode: FTMSControlPointOpCode, parameter: int = 0):
    if opcode == FTMSControlPointOpCode.REQUEST_CONTROL:
        return b"\x00"
    elif opcode == FTMSControlPointOpCode.RESET:
        return b"\x01"
    elif opcode == FTMSControlPointOpCode.SET_TARGET_SPEED:
        # parameter: uint16, 0.01km/h
        return b"\x02" + parameter.to_bytes(2, "little", signed=False)
    elif opcode == FTMSControlPointOpCode.SET_TARGET_INCLINE:
        # parameter: sint16, 0.1%
        return b"\x03" + parameter.to_bytes(2, "little", signed=True)
    elif opcode == FTMSControlPointOpCode.SET_TARGET_RESISTANCE_LEVEL:
        # parameter: uint8, 0.1 unitless
        return b"\x04" + parameter.to_bytes(1, "little", signed=False)
    elif opcode == FTMSControlPointOpCode.SET_TARGET_POWER:
        # parameter: sint16, 1W
        return b"\x05" + parameter.to_bytes(2, "little", signed=True)
    elif opcode == FTMSControlPointOpCode.START_OR_RESUME:
        # parameter: 01=stop, 02=pause
        return b"\x06"
    elif opcode == FTMSControlPointOpCode.STOP_OR_PAUSE:
        return b"\x07" + parameter.to_bytes(1, "little", signed=False)
    elif opcode == FTMSControlPointOpCode.RESPONSE_CODE:
        return b"\x80"
    else:
        raise ValueError("Invalid opcode")


ControlPointResponse = namedtuple("ControlPointResponse", ["request_code_enum", "result_code_enum"])


def parse_control_point_response(message: bytearray) -> ControlPointResponse:
    request_code_enum = FTMSControlPointOpCode(message[1])
    result_code_enum = FTMSControlPointResponseResultCode(message[2])
    return ControlPointResponse(request_code_enum, result_code_enum)
